//***************************************************************************
// Ces applets viennent en complment du livre :
//
//  		VIE ARTIFICIELLE
//      O la biologie rencontre l'informatique
//		Illustr en Java
//            Jean-Philippe Rennard
//           	  Vuibert 2002
//===========================================================================
//  L-systmes
//    v:0.1 09/09/2002
//
//  Cette applet permet de construire des L-systmes.
//
//    Jean-Philippe Rennard 2001-2002
//
//    This program is free software; you can redistribute it and/or
//    modify it under the terms of the GNU General Public License
//    as published by the Free Software Foundation.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//***************************************************************************
package lsys;

import java.awt.*;
import java.util.*;
import vabase.*;

/** Classe de gestion du L-systme */
public class Clsys extends Cprocess {
  static final int NB_REGLES=6;
  /** Pointeur sur l'interface utilisateur */
  private lsysUI appUI;
  /** Vue */
  vueLsys lsysVue;
  /** Tableau des rgles */
  Cregle regles[];
  /** Axiome */
  String axiome;
  /** Reprsentation actuelle=Gnome */
  String genome;
  /** Nombre d'itrations courant */
  int iteration;
  /** Nombre d'embranchements 8=45 degrs */
  int nbDir;
  /** Longueur de base d'un brin */
  float lgBase;
  /** Ratio de rduction d'chelle */
  float ratioEchelle;
  /** Angle de dpart */
  float angleDep;
  /** Position de dpart en x */
  float posDepX;
  /** Position de dpart en y */
  float posDepY;
  /** Pile de gestion de noeuds */
  private Stack pile;
  /** Noeud courant */
  private Cnoeud noeudCrt;
  /** Noeud de travail */
  private Cnoeud noeudW;
  /** Position courante dans le gnome */
  private int posGenome=0;
  /** Niveau courant d'embranchement */
  private int nivEmbranche;
  /** Flag d'affichage direct */
  private boolean affDirect;

  /** Construit un L-systme dans une interface utilisateur */
  public Clsys(lsysUI ui) {
    super(ui);
    appUI=ui;
    lsysVue=new vueLsys(this, Color.white, Color.black);
    vue=(Cvue)lsysVue;
    pile=new Stack();
    regles=new Cregle[NB_REGLES];
    noeudCrt=new Cnoeud(0,0,0,0);
    noeudW=new Cnoeud(0,0,0,0);
    affDirect=false;
    setTempo(1);
    nivEmbranche=0;
    choixRegle(0);
  }
  /** Mise  jour des rgles */
  public void setRegles(float rl, int dir, float lg, float ad, float px,
          float py, int it, String ax,
          char c1, String s1, char c2, String s2, char c3, String s3,
          char c4, String s4, char c5, String s5, char c6, String s6) {
    ratioEchelle=rl;
    nbDir=dir;
    lgBase=lg;
    angleDep=ad;
    posDepX=px;
    posDepY=py;
    setIteration(it);
    setAxiome(ax);
    regles[0]=new Cregle(c1,s1);
    regles[1]=new Cregle(c2,s2);
    regles[2]=new Cregle(c3,s3);
    regles[3]=new Cregle(c4,s4);
    regles[4]=new Cregle(c5,s5);
    regles[5]=new Cregle(c6,s6);
    noeudCrt=new Cnoeud(posDepX,posDepY,angleDep,0);
    initVue();
  }
  /** Mise  jour de l'axiome */
  public void setAxiome(String a) {
    axiome=a;
  }
  /** Mise  jour du nombre d'itrations */
  public void setIteration(int i) {
    iteration=i;
  }
  /** Application des rgles au gnome courant */
  public void appliqueRegles() {
    appUI.txtGenome.setText("Construction en cours...");
    genome=developpe(genome,iteration);
    appUI.txtGenome.setText(genome);
    appUI.txtLenGenome.setText(Integer.toString(genome.length()));
  }
  /** Procdure rcursive de dveloppement du gnome
  au niveau i  partir de l'axiome */
  private String developpe(String s, int i) {
    if(i==1)
      return substitue(axiome);
    return substitue(developpe(s,i-1));
  }
  /** Procdure de substitution des caractres par application des rgles */
  private String substitue(String s) {
    for(int j=0;j<s.length();j++) {
      for(int k=0;k<NB_REGLES;k++) {
        String gb=s;
        gb=regles[k].appliqueRegle(gb,j);
        if(gb!="") {
          j += gb.length()-s.length();
          s=gb;
          break;
        }
      }
    }
    return(s);
  }
  /** dmarrage du processus */
  public void lancement(boolean ad) {
    affDirect=ad;
    super.lancement();
  }

  /** Calcul de la gnration suivante */
  public void geneSuiv() {
    if(posGenome==genome.length()) {
      lsysVue.repaint();
      stop();
      appUI.stop();
      initParcours();
      return;
    }
    char symbCrt=genome.charAt(posGenome);
    switch(symbCrt) {
      case 'F':
        float lg=(float)Math.pow(ratioEchelle,nivEmbranche)*lgBase;
        float posX=noeudCrt.posX+(float)(Math.cos(noeudCrt.dirCrt*
                                              ((2.0*Math.PI)/360.0))*lg);
        float posY=noeudCrt.posY+(float)(Math.sin(noeudCrt.dirCrt*
                                              ((2.0*Math.PI)/360.0))*lg);
        lsysVue.affiche(noeudCrt, posX, posY);
        noeudCrt.posX=posX;
        noeudCrt.posY=posY;
      break;
      case 'f':
        lg=(float)Math.pow(ratioEchelle,nivEmbranche)*lgBase;
        posX=noeudCrt.posX+(float)(Math.cos(noeudCrt.dirCrt*
                                              ((2.0*Math.PI)/360.0))*lg);
        posY=noeudCrt.posY+(float)(Math.sin(noeudCrt.dirCrt*
                                              ((2.0*Math.PI)/360.0))*lg);
        noeudCrt.posX=posX;
        noeudCrt.posY=posY;
        noeudCrt.colCrt=0;
      break;
      case '+':
        noeudCrt.dirCrt += 360.0 /nbDir;
      break;
      case '-':
        noeudCrt.dirCrt -= 360.0 /nbDir;
      break;
      case 'c':
        noeudCrt.colCrt++;
      break;
      case '[':
        nivEmbranche++;
        pile.push(new Cnoeud(noeudCrt));
      break;
      case ']':
        nivEmbranche--;
        noeudCrt=(Cnoeud)pile.pop();
      break;
    }
    generation++;
    posGenome++;
    appUI.affGeneration();
    if(!affDirect) lsysVue.repaint();
  }
  /** Initialisation de la vue */
  public void initVue() {
    initParcours();
    lsysVue.setNewTaille();
  }
  /** retour au dpart de la construction */
  public void initParcours() {
    posGenome=0;
    nivEmbranche=0;
    generation=0;
    noeudCrt=new Cnoeud(posDepX,posDepY,angleDep,0);
  }

  /** Slection d'une rgle existante */
  // Plusieurs rgles viennent de LSysEdit de John Martin Caroll,
  // http://www.geocities.com/siliconvalley/campus/1519
  public void choixRegle(int n) {
    switch(n) {
      case 0:
        // Plante 1
        ratioEchelle=0.75F;
        nbDir=24;
        lgBase=7.0F;
        angleDep=90.0F;
        posDepX=500.0F;
        posDepY=0.0F;
        setIteration(6);
        setAxiome("F");
        regles[0]=new Cregle('F',"FF[c+F][c-F]");
        regles[1]=new Cregle('_',"");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 1:
        // Plante 2
        ratioEchelle=1.0F;
        nbDir=17;
        lgBase=3.0F;
        angleDep=90.0F;
        posDepX=500.0F;
        posDepY=0.0F;
        setIteration(7);
        setAxiome("X");
        regles[0]=new Cregle('X',"F[c+X][c-X]FX");
        regles[1]=new Cregle('F',"FF");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 2:
        // Koch
        ratioEchelle=1.0F;
        nbDir=5;
        lgBase=3.0F;
        angleDep=0.0F;
        posDepX=300.0F;
        posDepY=200.0F;
        setIteration(5);
        setAxiome("F+F+F+F+F");
        regles[0]=new Cregle('F',"F-F++F-F");
        regles[1]=new Cregle('_',"");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 3:
        // Buisson 1
        ratioEchelle=1.0F;
        nbDir=13;
        lgBase=10.0F;
        angleDep=90.0F;
        posDepX=500.0F;
        posDepY=0.0F;
        setIteration(5);
        setAxiome("G");
        regles[0]=new Cregle('G',"GfX[+G][-G]");
        regles[1]=new Cregle('X',"X[-FFF][+FFF]FX");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 4:
        // Buisson 2
        ratioEchelle=1.0F;
        nbDir=15;
        lgBase=10.0F;
        angleDep=90.0F;
        posDepX=500.0F;
        posDepY=100.0F;
        setIteration(5);
        setAxiome("fF");
        regles[0]=new Cregle('F',"fF+[c+F-F-F]-[c-fF+F+F]");
        regles[1]=new Cregle('_',"");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 5:
        // Triangles
        ratioEchelle=1.0F;
        nbDir=3;
        lgBase=15.0F;
        angleDep=0.0F;
        posDepX=500.0F;
        posDepY=100.0F;
        setIteration(7);
        setAxiome("fX");
        regles[0]=new Cregle('X',"--FXF++FXF++FXF--");
        regles[1]=new Cregle('_',"");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 6:
        // Squigly
        ratioEchelle=1.0F;
        nbDir=3;
        lgBase=7.0F;
        angleDep=0.0F;
        posDepX=300.0F;
        posDepY=300.0F;
        setIteration(7);
        setAxiome("r");
        regles[0]=new Cregle('r',"Fr+FL+Fr");
        regles[1]=new Cregle('L',"FL-Fr-FL");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 7:
        // Hilbert
        ratioEchelle=1.0F;
        nbDir=4;
        lgBase=10.0F;
        angleDep=90.0F;
        posDepX=200.0F;
        posDepY=100.0F;
        setIteration(6);
        setAxiome("X");
        regles[0]=new Cregle('X',"-YF+XFX+FY-");
        regles[1]=new Cregle('Y',"+XF-YFY-FX+");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 8:
        // plantes
        ratioEchelle=1.0F;
        nbDir=18;
        lgBase=5.0F;
        angleDep=180.0F;
        posDepX=750.0F;
        posDepY=0.0F;
        setIteration(5);
        setAxiome("[---X]f[----X]f[----X]f[-----X]");
        regles[0]=new Cregle('X',"F[c+X]F[c-X]+X");
        regles[1]=new Cregle('F',"FF");
        regles[2]=new Cregle('f',"ff");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 9:
        // Sierpinski
        ratioEchelle=1.0F;
        nbDir=3;
        lgBase=3.0F;
        angleDep=0.0F;
        posDepX=150.0F;
        posDepY=0.0F;
        setIteration(7);
        setAxiome("FXF--FF--FF");
        regles[0]=new Cregle('X',"--FXF++FXF++FXF--");
        regles[1]=new Cregle('F',"FF");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 10:
        // Arbre
        ratioEchelle=1.2F;
        nbDir=24;
        lgBase=3.5F;
        angleDep=90.0F;
        posDepX=500.0F;
        posDepY=0.0F;
        setIteration(7);
        setAxiome("G");
        regles[0]=new Cregle('G',"GfX[+G][-G]");
        regles[1]=new Cregle('X',"cX[-FFF][+FFF]FX");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      case 11:
        // Plante 3
        ratioEchelle=0.5F;
        nbDir=12;
        lgBase=10.0F;
        angleDep=90.0F;
        posDepX=500.0F;
        posDepY=0.0F;
        setIteration(4);
        setAxiome("F");
        regles[0]=new Cregle('F',"cF[+F[+F][-F]F][-F[+F][-F]F]F[+F][-F]F");
        regles[1]=new Cregle('_',"");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
      // ABOP 1.24
      case 12:
      default:
        ratioEchelle=1.0F;
        nbDir=32;
        lgBase=8.0F;
        angleDep=90.0F;
        posDepX=200.0F;
        posDepY=0.0F;
        setIteration(5);
        setAxiome("X");
        regles[0]=new Cregle('X',"F-[cXX+X]+F[c+FX]-X");
        regles[1]=new Cregle('F',"FF");
        regles[2]=new Cregle('_',"");
        regles[3]=new Cregle('_',"");
        regles[4]=new Cregle('_',"");
        regles[5]=new Cregle('_',"");
      break;
    }
    noeudCrt=new Cnoeud(posDepX,posDepY,angleDep,0);
  }
}
